/*
 * Copyright 2008 by Kappich Systemberatung, Aachen
 * Copyright 2007 by Kappich Systemberatung, Aachen
 * Copyright 2006 by Kappich Systemberatung Aachen
 * Copyright 2006 by Kappich+Kni Systemberatung Aachen (K2S)
 * 
 * This file is part of de.bsvrz.puk.config.
 * 
 * de.bsvrz.puk.config is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.puk.config is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.puk.config; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.puk.config.configFile.datamodel;

import de.bsvrz.dav.daf.communication.dataRepresentation.AttributeBaseValueDataFactory;
import de.bsvrz.dav.daf.communication.dataRepresentation.AttributeHelper;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataAndATGUsageInformation;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.puk.config.configFile.fileaccess.*;
import de.bsvrz.puk.config.main.managementfile.VersionInfo;
import de.bsvrz.puk.config.xmlFile.properties.ConfigurationAreaChangeInformation;
import de.bsvrz.sys.funclib.dataSerializer.Deserializer;
import de.bsvrz.sys.funclib.dataSerializer.Serializer;
import de.bsvrz.sys.funclib.dataSerializer.SerializingFactory;
import de.bsvrz.sys.funclib.debug.Debug;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.util.*;

/**
 * Implementierung des Konfigurationsbereichs auf Seiten der Konfiguration.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 13141 $
 */
public class ConfigConfigurationArea extends ConfigConfigurationObject implements ConfigurationArea, ConfigConfigurationAreaInterface {

	/** DebugLogger fr Debug-Ausgaben */
	private static final Debug _debug = Debug.getLogger();

	/** Eine Referenz auf das Datenmodell. */
	private final ConfigDataModel _dataModel;

	/** Konfigurationsverantwortlicher dieses Konfigurationsbereichs. */
	private ConfigurationAuthority _configurationAuthority = null;

	/**
	 * Diese Liste enthlt alle Abhngigkeiten des Bereichs. Die Abhngigkeiten sind dabei in der Reihenfolge ihres Auftretens gespeichert. Der letzte Eintrag ist
	 * somit die letzte Abhngigkeite, die eingetragen wurde (und somit die aktuellste).
	 */
	private final List<ConfigurationAreaDependency> _areaDependencyList = new ArrayList<ConfigurationAreaDependency>();

	/**
	 * Mit dieser Verwendung kann ein Datensatz engefordert werden, der die Abhngigkeiten dieses Bereichs enthlt. Ist der Wert <code>null</code>, so ist die
	 * bentigte ATG nicht bekannt und der Mechanismus wird nicht benutzt.
	 */
	private final AttributeGroupUsage _atguForDependencies;

	/** Mit diesem Enum kann angegeben werden, ob ein dynamisches Objekt, ein Konfigurationsobjekt oder ein konfigurierender Datensatz gendert wurde. */
	enum KindOfLastChange {

		DynamicObject,
		ConfigurationObject,
		ConfigurationData;
	}

	/**
	 * Konstruktor eines System-Objekt fr einen Konfigurationsbereich.
	 *
	 * @param dataModel        das Datenmodell, welches alle Konfigurationsbereiche enthlt
	 * @param systemObjectInfo das korrespondierende Objekt aus den Konfigurationsdateien
	 */
	public ConfigConfigurationArea(ConfigDataModel dataModel, SystemObjectInformationInterface systemObjectInfo) {
		super(null, systemObjectInfo);
		_dataModel = dataModel;

		final AttributeGroup attributeGroup = _dataModel.getAttributeGroup("atg.konfigurationsBereichAbhngigkeiten");
		if(attributeGroup != null) {
			_atguForDependencies = attributeGroup.getAttributeGroupUsage(
					_dataModel.getAspect(
							"asp.eigenschaften"
					)
			);
			loadDependencyDataSets();
		}
		else {
			// Die Konfiguration untersttzt die ATG nicht
			_atguForDependencies = null;
		}
	}

	public ConfigDataModel getDataModel() {
		return _dataModel;
	}

	public ConfigConfigurationArea getConfigurationArea() {
		return this;
	}

	public synchronized ConfigurationAuthority getConfigurationAuthority() {
		// Der KV wird hier gecached, damit sichergestellt ist, dass sich der KV zur Laufzeit niemals ndert. Erst bei einem Neustart der Konfiguration
		// kann er sich ndern.
		if(_configurationAuthority == null) {
			final AttributeGroup atg = getDataModel().getAttributeGroup("atg.konfigurationsBereichEigenschaften");
			if(atg == null) throw new IllegalStateException("Attributgruppe atg.konfigurationsBereichEigenschaften wurde nicht gefunden");
			final Aspect asp = getDataModel().getAspect("asp.eigenschaften");
			if(asp == null) throw new IllegalStateException("Aspekt asp.eigenschaften wurde nicht gefunden");
			final Data data = getConfigurationData(atg, asp);
			if(data == null) {
				throw new IllegalStateException(
						"Der Datensatz mit dem Konfigurationsverantwortlichen des Konfigurationsbereichs " + getNameOrPidOrId()
						+ " konnte nicht gelesen werden."
				);
			}
			_configurationAuthority = (ConfigurationAuthority)data.getReferenceValue("zustndiger").getSystemObject();
			if(_configurationAuthority == null) {
				throw new IllegalStateException(
						"Der Konfigurationsverantwortliche des Konfigurationsbereichs " + getNameOrPidOrId() + " konnte nicht ermittelt werden."
				);
			}
		}
		return _configurationAuthority;
	}

	@Override
	public void setConfigurationData(final AttributeGroupUsage atgUsage, final Data data) throws ConfigurationChangeException {
		// Verhindern, dass der Parameter des Konfigurationsverantwortlichen gendert wird.
		final AttributeGroupUsage usage = getDataModel().getAttributeGroup("atg.konfigurationsBereichEigenschaften").getAttributeGroupUsage(
				getDataModel().getAspect("asp.eigenschaften")
		);
		if(atgUsage.equals(usage)){
			if(!data.getReferenceValue("zustndiger").getSystemObject().equals(getConfigurationAuthority())){
				throw new ConfigurationChangeException("Der Konfigurationsverantwortliche eines Konfigurationsbereichs kann nicht gendert werden.");
			}
		}

		// Normale Funktion aufrufen
		super.setConfigurationData(atgUsage, data);
	}

	/**
	 * Wird whrend der Aktivierung aufgerufen um den neuen KV zu setzen. Der Weg ber setConfigurationData() ist nicht mglich,
	 * da dort das Setzen des KV eine Exception produziert.
	 * @param newAuthority
	 * @throws ConfigurationChangeException
	 */
	public void activateNewAuthority(final SystemObject newAuthority) throws ConfigurationChangeException {
		final AttributeGroup configurationAreaAtg = getDataModel().getAttributeGroup("atg.konfigurationsBereichEigenschaften");
		final Aspect configurationAreaAsp = getDataModel().getAspect("asp.eigenschaften");

		final Data data = getConfigurationData(configurationAreaAtg, configurationAreaAsp);
		data.getReferenceValue("zustndiger").setSystemObject(newAuthority); // Zustndigen setzen
		data.getReferenceValue("neuerZustndiger").setSystemObject(null);    // neuen Zustndigen lschen
		
		super.setConfigurationData(configurationAreaAtg.getAttributeGroupUsage(configurationAreaAsp), data);   // Datensatz speichern
	}

	public short getActivatableVersion() {
		Data data = getConfigurationData(
				getDataModel().getAttributeGroup("atg.konfigurationsBereichbernahmeInformationen"), getDataModel().getAspect("asp.eigenschaften")
		);
		if(data != null) {
			return data.getUnscaledValue("aktivierbareVersion").shortValue();
		}
		else {
			throw new IllegalStateException("Die aktivierbare Version des Konfigurationsbereichs " + getNameOrPidOrId() + " konnte nicht ermittelt werden.");
		}
	}

	/**
	 * Setzt die Version, die zur bernahme und Aktivierung dieses Konfigurationsbereichs in anderen Konfigurationen freigegeben werden soll.
	 *
	 * @param activatableVersion die Versionsnummer, die zur bernahme und Aktivierung dieses Konfigurationsbereichs freigegeben werden soll
	 *
	 * @throws ConfigurationChangeException Falls die Versionsnummer nicht am Konfigurationsbereich gespeichert werden kann.
	 */
	void setActivatableVersion(short activatableVersion) throws ConfigurationChangeException {
		final AttributeGroup atg = getDataModel().getAttributeGroup("atg.konfigurationsBereichbernahmeInformationen");
		Data data = getConfigurationData(atg);
		// Datensatz muss vorhanden sein, da ein neuer Bereich immer mit diesem Datensatz erzeugt wird.
		if(data != null) {
			data.getUnscaledValue("aktivierbareVersion").set(activatableVersion);
			setConfigurationData(atg, data);
		}
		else {
			throw new ConfigurationChangeException(
					"Die Version, die zur bernahme und Aktivierung freigegeben werden soll, konnte nicht geschrieben werden, da der Datensatz nicht vorhanden ist."
			);
		}
	}

	public short getTransferableVersion() {
		Data data = getConfigurationData(
				getDataModel().getAttributeGroup("atg.konfigurationsBereichbernahmeInformationen"), getDataModel().getAspect("asp.eigenschaften")
		);
		if(data != null) {
			return data.getUnscaledValue("bernehmbareVersion").shortValue();
		}
		else {
			throw new IllegalStateException("Die bernehmbare Version des Konfigurationsbereichs " + getNameOrPidOrId() + " konnte nicht ermittelt werden.");
		}
	}

	/**
	 * Setzt die Version, die zur bernahme dieses Konfigurationsbereichs in anderen Konfigurationen freigegeben werden soll.
	 *
	 * @param transferableVersion die Versionsnummer, die zur bernahme dieses Konfigurationsbereichs freigegeben werden soll
	 *
	 * @throws ConfigurationChangeException Falls die Versionsnummer nicht am Konfigurationsbereich gespeichert werden kann.
	 */
	void setTransferableVersion(short transferableVersion) throws ConfigurationChangeException {
		final AttributeGroup atg = getDataModel().getAttributeGroup("atg.konfigurationsBereichbernahmeInformationen");
		Data data = getConfigurationData(atg);
		// Datensatz muss vorhanden sein, da ein neuer Bereich immer mit diesem Datensatz erzeugt wird.
		if(data != null) {
			data.getUnscaledValue("bernehmbareVersion").set(transferableVersion);
			setConfigurationData(atg, data);
		}
		else {
			throw new ConfigurationChangeException(
					"Die Version, die zur bernahme freigegeben werden soll, konnte nicht geschrieben werden, da der Datensatz nicht vorhanden ist."
			);
		}
	}

	public short getActiveVersion() {
		return _dataModel.getActiveVersion(getConfigurationArea());
	}

	public short getModifiableVersion() {
		// vernderbare Version
		final ConfigurationAreaFile areaFile = getDataModel().getConfigurationFileManager().getAreaFile(getPid());
		return areaFile.getNextActiveVersion();
	}

	/**
	 * Gibt die in Bearbeitung befindliche Version zurck, wenn nderungen in dieser Version dieses Bereichs stattgefunden haben. nderungen knnen sein: <ul>
	 * <li>Objekt wurde auf ungltig gesetzt</li> <li>Objekt wurde erstellt</li> <li>Elemente wurden einer Menge hinzugefgt</li> <li>Elemente wurden aus einer
	 * Menge entfernt</li> </ul>
	 * <p/>
	 * Wurden keine nderungen in der in Bearbeitung befindlichen Version vorgenommen, wird die Version davor zurckgegeben.
	 *
	 * @return Falls nderungen in der in Bearbeitung befindlichen Version durchgefhrt wurden, wird diese Version zurckgegeben, sonst die Version davor.
	 */
	public short getLastModifiedVersion() {
		final short modifiableVersion = getModifiableVersion();

		// Hier wird geprft, ob ein aktuelles Objekt in der "modifiableVersion" auf ungltig gesetzt wurde.
		// Ist dies der Fall, wird die "modifiableVersion" zurckgegeben.
		final Collection<SystemObject> currentObjects = getCurrentObjects();
		for(SystemObject currentObject : currentObjects) {
			// Dynamische Objekte mssen nicht geprft werden, da sie sofort gltig sind
			if(currentObject instanceof ConfigurationObject) {
				final ConfigurationObject configurationObject = (ConfigurationObject)currentObject;
				if(configurationObject.getNotValidSince() == modifiableVersion) {
					// dieses Objekt wurde auf ungltig gesetzt
					return modifiableVersion;
				}
			}
		}
		// Kein aktuelles Objekt wurde in der "modifiableVersion" auf ungltig gesetzt

		// Hier werden die zuknftig aktuellen Objekte geprft, ob sie in der "modifiableVersion auf ungltig gesetzt oder erstellt wurde.
		// Ist dies der Fall, wird die "modifiableVersion" zurckgegeben.
		final Collection<SystemObject> newObjects = getNewObjects();
		for(SystemObject newObject : newObjects) {
			// da die Methode getNewObjects() nur KonfigurationsObjekte zurckgibt, muss hier nicht geprft werden
			final ConfigurationObject configurationObject = (ConfigurationObject)newObject;
			if(configurationObject.getNotValidSince() == modifiableVersion || configurationObject.getValidSince() == modifiableVersion) {
				// es gibt zuknftige Objekte, die in der "modifiableVersion" ungltig werden
				// es gibt neu erstellte Objekte in der Version "modifiableVersion"
				return modifiableVersion;
			}
		}

		// Prft die Mengen mit Referenzierungsart "Assoziation" dieses Bereichs, ob Elemente in
		// der Version "modifiableVersion" ungltig oder neu hinzugefgt wurden/werden.
		final Collection<SystemObjectType> objectTypes = new LinkedList<SystemObjectType>();
		objectTypes.add(getDataModel().getType("typ.konfigurationsMenge"));
		final Collection<SystemObject> systemObjects = getObjects(objectTypes, ObjectTimeSpecification.valid());
		for(SystemObject systemObject : systemObjects) {
			if(systemObject instanceof ConfigNonMutableSet) {
				final ConfigNonMutableSet nonMutableSet = (ConfigNonMutableSet)systemObject;
				// nur versionierte Mengen mssen betrachtet werden
				if(nonMutableSet.getObjectSetType().getReferenceType() == ReferenceType.ASSOCIATION) {
					// dieser Menge wurde in der "modifiableVersion" mindestens ein Element hinzugefgt oder entfernt
					if(nonMutableSet.isSetChanged(modifiableVersion)) return modifiableVersion;
				}
			}
			else {
				_debug.error("Element vom Typ 'typ.konfigurationsMenge' ist kein konfigurierender MengenTyp: ", systemObject);
				throw new IllegalStateException("Element vom Typ 'typ.konfigurationsMenge' ist kein konfigurierender MengenTyp: " + systemObject.getPid());
			}
		}

		// Prfung auf unversionierte nderungen
		AttributeGroup attributeGroup = _dataModel.getAttributeGroup("atg.konfigurationsBereichUnversioniertenderungen");
		if(attributeGroup != null){
			Data configurationData = getConfigurationData(attributeGroup);
			if(configurationData != null){
				for(Data data : configurationData.getItem("versionen")) {
					if(data.getUnscaledValue("Version").shortValue() == modifiableVersion) return modifiableVersion;
				}
			}
		}

		// Weder ein aktuelles noch ein zuknftiges Objekt oder eine Menge wurde in der "modifiableVersion" gendert.
		// Aus diese Grund, wird die "modifiableVersion" - 1 zurckgegeben.
		// MAX(ActiveVersion, ActivatableVersion, TransferableVersion) = ModifiableVersion - 1;
		return (short)(modifiableVersion - 1);
	}

	public long getTimeOfLastDynamicChange() {
		Data data = getConfigurationData(getDataModel().getAttributeGroup("atg.konfigurationsBereichnderungsZeiten"));
		if(data != null) {
			return data.getTimeValue("LetztenderungszeitDynamischesObjekt").getMillis();
		}
		else {
			return 0;
		}
	}

	public long getTimeOfLastNonActiveConfigurationChange() {
		Data data = getConfigurationData(getDataModel().getAttributeGroup("atg.konfigurationsBereichnderungsZeiten"));
		if(data != null) {
			return data.getTimeValue("LetztenderungszeitKonfigurationsObjekt").getMillis();
		}
		else {
			return 0;
		}
	}

	public long getTimeOfLastActiveConfigurationChange() {
		Data data = getConfigurationData(getDataModel().getAttributeGroup("atg.konfigurationsBereichnderungsZeiten"));
		if(data != null) {
			return data.getTimeValue("LetztenderungszeitDatensatz").getMillis();
		}
		else {
			return 0;
		}
	}

	public Collection<SystemObject> getObjects(Collection<SystemObjectType> systemObjectTypes, ObjectTimeSpecification timeSpecification) {
		// alle Objekt-Typen, die zu betrachten sind, ermitteln - also auch die Typen, die die angegebenen Typen erweitern
		final Set<SystemObjectType> relevantObjectTypes = new HashSet<SystemObjectType>();
		if(systemObjectTypes == null) {
			// es wurde keine Einschrnkung der Typen gegeben -> alle Typen ermitteln
			// die Typen eines TypTyp-Objekts ermittelt man mit getElements()!
			relevantObjectTypes.add(getDataModel().getTypeTypeObject());
			final List<SystemObject> typeTypeElements = getDataModel().getTypeTypeObject().getElements();
			for(SystemObject typeTypeElement : typeTypeElements) {
				relevantObjectTypes.add((SystemObjectType)typeTypeElement);
			}
		}
		else {
			// die Typen eines Typ-Objekts werden ber die Sub-Types ermittelt.
			for(SystemObjectType objectType : systemObjectTypes) {
				relevantObjectTypes.add(objectType);
				relevantObjectTypes.addAll(getDataModel().getAllSubTypes(objectType));
			}
		}
		return getDirectObjects(relevantObjectTypes, timeSpecification);
	}

	public Collection<SystemObject> getDirectObjects(Collection<SystemObjectType> systemObjectTypes, ObjectTimeSpecification timeSpecification) {
		// Parameter null ist hier nicht erlaubt
		if(systemObjectTypes == null || timeSpecification == null) {
			throw new IllegalArgumentException("Parameter 'null' ist hier nicht erlaubt.");
		}

		// speichert alle IDs der zu betrachtenden Typen ab
		final Collection<Long> typeIds = new ArrayList<Long>();
		// Typ-IDs ermitteln
		for(SystemObjectType objectType : systemObjectTypes) {
			typeIds.add(objectType.getId());
		}

		// Konfigurationsbereichsdatei ermitteln
		ConfigurationAreaFile areaFile = getDataModel().getConfigurationFileManager().getAreaFile(getConfigurationArea().getPid());

		SystemObjectInformationInterface[] systemObjectInfos;

		// Anfangs- und Endzeit ermitteln
		long startTime;
		long endTime;
		if(timeSpecification.getType() == TimeSpecificationType.VALID) {
			systemObjectInfos = areaFile.getActualObjects(typeIds);
		}
		else {
			if(timeSpecification.getType() == TimeSpecificationType.VALID_AT_TIME) {
				startTime = timeSpecification.getTime();
				endTime = timeSpecification.getTime();
			}
			else {
				startTime = timeSpecification.getStartTime();
				endTime = timeSpecification.getEndTime();
			}

			// lokale Zeit fr aktivierte Versionen (LOCAL_ACTIVATION_TIME)
			systemObjectInfos = areaFile.getObjects(
					startTime, endTime, ConfigurationAreaTime.LOCAL_ACTIVATION_TIME, timeSpecification.getType(), typeIds
			);
		}

		// SystemObjekte holen
		final Collection<SystemObject> objects = new ArrayList<SystemObject>();
		for(SystemObjectInformationInterface systemObjectInfo : systemObjectInfos) {
			SystemObject systemObject = getDataModel().createSystemObject(systemObjectInfo);
			objects.add(systemObject);
		}
		return Collections.unmodifiableCollection(objects);
	}

	public Collection<SystemObject> getCurrentObjects() {
		return getDataModel().getCurrentObjects(getConfigurationArea());
	}

	public Collection<SystemObject> getNewObjects() {
		return getDataModel().getNewObjects(getConfigurationArea());
	}

	public ConfigurationObject createConfigurationObject(ConfigurationObjectType type, String pid, String name, Collection<? extends ObjectSet> sets)
			throws ConfigurationChangeException {
		// wenn bei der pid oder dem namen "null" bergeben wird, wird der Leerstring ("") verwendet.
		if(pid == null) pid = "";
		if(name == null) name = "";
		// 1. prfen, ob berhaupt gendert werden darf
		{
			if(checkChangePermit()) {

				// Namen und Pids drfen nur eine Lnge von 255 Zeichen besitzen
				checkNameLength(name);
				checkPidLength(pid);

				// gibt es bereits ein Objekt mit der angegebenen Pid in diesem Bereich, so wird kein neues Objekt angelegt, sondern der Vorgang mit einer
				// ConfigurationChangeException abgebrochen
				final ConfigDataModel configDataModel = getDataModel();
				final ConfigurationFileManager fileManager = configDataModel.getConfigurationFileManager();

				try {
					short modifiableVersion = getModifiableVersion();
					short transferableVersion = getTransferableVersion();

					// wurde keine Pid angegeben, so muss auch nicht berprft werden (z.B. beim Erstellen von Mengen)
					if(!pid.equals("")) {
						SystemObjectInformationInterface[] systemObjectInfos = fileManager.getNewObjects(pid);
						if(systemObjectInfos.length > 0) {
							// zuknftige Elemente berprfen
							for(SystemObjectInformationInterface systemObjectInfo : systemObjectInfos) {
								ConfigurationObjectInfo configurationObjectInfo = (ConfigurationObjectInfo)systemObjectInfo;
								if(configurationObjectInfo.getFirstValidVersion() > transferableVersion && (
										configurationObjectInfo.getFirstInvalidVersion() == 0
										|| configurationObjectInfo.getFirstInvalidVersion() > modifiableVersion)) {
									throw new IllegalStateException(
											"Es existiert bereits ein zuknftiges Objekt mit gleicher Pid '" + pid + "' in diesem Bereich "
											+ getConfigurationArea() + ": " + systemObjectInfo
									);
								}
							}
						}
					}
					// ein zuknftiges kollidierendes Objekt existiert nicht
					// aktuell gltiges Objekt berpfen
					SystemObject existAnotherObject = getDataModel().getObject(pid);
					if(existAnotherObject != null && existAnotherObject.getConfigurationArea() == getConfigurationArea()
					   && !existAnotherObject.getType().isConfigurating()) {
						throw new IllegalStateException(
								"Es existiert bereits ein aktuelles dynamisches Objekt mit gleicher Pid '" + pid + "' in diesem Bereich "
								+ getConfigurationArea()
						);
					}
				}
				catch(IllegalStateException ex) {
					// neue ConfigurationChangeException werfen
					final String errorMessage = "Es konnte kein neues konfigurierendes Objekt angelegt werden";
					_debug.error(errorMessage, ex.getMessage());
					throw new ConfigurationChangeException(errorMessage, ex);
				}

				// 2. nchste freie ID ermitteln
				long nextObjectId = configDataModel.getNextObjectId();
				final ConfigurationAreaFile areaFile = fileManager.getAreaFile(getPid());
				final ConfigurationObjectInfo objectInfo = areaFile.createConfigurationObject(nextObjectId, type.getId(), pid, name);
				setTimeOfLastChanges(KindOfLastChange.ConfigurationObject);
				// die Mengen werden in einem konfigurierenden Datensatz mit der Attributgruppenverwendungs-ID "-1" abgespeichert
				if(sets != null) createConfigurationDataForSets(objectInfo, sets, getSerializerVersion());

				final ConfigurationObject configObject = (ConfigurationObject)configDataModel.createSystemObject(objectInfo);
				_debug.fine("Neues konfigurierendes Objekt angelegt: " + configObject.getPidOrNameOrId() + " Id: " + configObject.getId());
				return configObject;
			}
			else {
				throw new ConfigurationChangeException(
						"Fr diesen Konfigurationsbereich " + getNameOrPidOrId()
						+ " ist die Konfiguration nicht berechtigt ein neues KonfigurationsObjekt anzulegen."
				);
			}
		}
	}

	public DynamicObject createDynamicObject(DynamicObjectType type, String pid, String name) throws ConfigurationChangeException {
		return createDynamicObject(type, pid, name, null, (short)0, false);
	}

	/**
	 * Erzeugt ein neues dynamisches System-Objekt eines vorgegebenen Typs mit einer angegebenen PID. Optional kann der Name des neuen Objekts vorgegeben werden.
	 * Das neue Objekt wird sofort gltig.
	 *
	 * @param type              Typ des neuen Objekts
	 * @param pid               PID des neuen Objekts. Der leere String ("") oder <code>null</code> wird als "keine PID" interpretiert.
	 * @param name              Name des neuen Objekts (kann spter verndert werden)
	 * @param simulationVariant Simulationsvariante des neuen Objekts
	 *
	 * @return Stellvertreterobjekt fr das neu angelegte dynamische Objekt.
	 *
	 * @throws ConfigurationChangeException Wenn das Objekt nicht erzeugt werden konnte.
	 */
	public DynamicObject createDynamicObject(DynamicObjectType type, String pid, String name, short simulationVariant) throws ConfigurationChangeException {
		return createDynamicObject(type, pid, name, null, simulationVariant, false);
	}

	public DynamicObject createDynamicObject(DynamicObjectType type, String pid, String name, Collection<DataAndATGUsageInformation> data)
			throws ConfigurationChangeException {
		return createDynamicObject(type, pid, name, data, (short)0, true);
	}

	public Collection<ConfigurationAreaDependency> getDependencyFromOtherConfigurationAreas() {
		final List<ConfigurationAreaDependency> copy = new ArrayList<ConfigurationAreaDependency>();
		copy.addAll(_areaDependencyList);
		return copy;
	}

	/**
	 * Diese Methode prft, ob eine Abhngigkeit gegenber einer anderen Abhngigkeite neuere Informationen enthlt, bei diesem Vergleich wird die Art der
	 * Abhngigkeit nicht bercksichtigt. Ein Vergleich ist nur dann sinnvoll, wenn beide Abhngigkeiten den selben Bereich referenzieren.
	 * <p/>
	 * <p/>
	 * Definition: Eine Abhngigkeit(A_neu) enthlt gegenber einer anderen Abhngigkeit(A_alt) neuere Informationen, wenn:<br> <ul> <li>1) Die Version in der die
	 * Abhngigkeite entdeckt wurde in A_Neu grer ist, als die der Abhngigkeit A_alt. Zustzlich muss noch die Version, in der der abhngige Bereich bentigt
	 * wird, grer sein als die aktuell gespeicherte Version. Oder aber A_alt war optional und A_neu ist notwendig. </li> <li>2) Sind die in 1) genannten
	 * Versionen gleich, so wird dann eine neuere Information gefunden, wenn die Version, in der der Abhngige Bereich bentigt wird, von A_neu grer ist als die
	 * Version, die in A_alt gespeichert ist.</li> <li></li> <li></li> </ul>
	 *
	 * @param oldAreaDependency   Abhngigkeit mit vermeindlich lteren Informationen
	 * @param newerAreaDependency Abhngigkeit mit vermeindlich neueren Informationen als <code>oldAreaDependency</code>
	 *
	 * @return <code>true</code>, wenn newerAreaDependency neuere Informationen enthlt als <code>oldAreaDependency</code>; <code>false</code>, sonst.
	 */
	private boolean hasNewVersionInformation(ConfigurationAreaDependency oldAreaDependency, ConfigurationAreaDependency newerAreaDependency) {
		if(newerAreaDependency.getDependencyOccurredAtVersion() > oldAreaDependency.getDependencyOccurredAtVersion()) {
			// 1)
			if(newerAreaDependency.getNeededVersion() > oldAreaDependency.getNeededVersion()) {
				return true;
			}
			else if(hasNewStateInformation(oldAreaDependency, newerAreaDependency) == true) {
				return true;
			}
			else {
				return false;
			}
		}
		else {
			// 2
			if(newerAreaDependency.getDependencyOccurredAtVersion() == oldAreaDependency.getDependencyOccurredAtVersion()) {
				if(newerAreaDependency.getNeededVersion() > oldAreaDependency.getNeededVersion()) {
					return true;
				}
				else {
					return false;
				}
			}
			else {
				// Die Version, in der die Abhngigkeit entdeckt wurde ist kleiner. Also kann sie keine neueren Infos besitzen.
				// 2)
				return false;
			}
		}
	}

	/**
	 * Prft, ob die Art der Abhngigkeite eine neue Information enthlt. Definition: Eine neue Abhngigkeite A_neu enthlt dann eine neuere Information als eine
	 * alte Abhngigkeite A_alt wenn: <ul> <li>1) Die Abhngigkeite A_neu notwendig ist und die alte Abhngigkeit A_alt optional</li> </ul>
	 *
	 * @param oldAreaDependency   Abhngigkeit, die bisher gespeichert wurde.
	 * @param newerAreaDependency Abhngigkeit, die vielleicht gespeichert werden soll
	 *
	 * @return <code>true</code>, wenn die neue Abhngigkeite neue Informationen enthlt; <code>false</code>, sonst.
	 */
	private boolean hasNewStateInformation(ConfigurationAreaDependency oldAreaDependency, ConfigurationAreaDependency newerAreaDependency) {
		if(newerAreaDependency.getKind() == ConfigurationAreaDependencyKind.REQUIRED
		   && oldAreaDependency.getKind() == ConfigurationAreaDependencyKind.OPTIONAL) {
			return true;
		}
		else {
			return false;
		}
	}

	/**
	 * Erzeugt eine neue Abhngigkeit die "notwendig" ist.
	 *
	 * @param areaDependence Abhngigkeit, die kopiert wird.
	 *
	 * @return Kopie von <code>areaDependence</code> mit der Eigenschaft notwendig.
	 */
	private ConfigurationAreaDependency createRequiredDependence(ConfigurationAreaDependency areaDependence) {
		return new ConfigurationAreaDependency(
				areaDependence.getDependencyOccurredAtVersion(),
				areaDependence.getNeededVersion(),
				areaDependence.getDependantArea(),
				ConfigurationAreaDependencyKind.REQUIRED
		);
	}

	/**
	 * Speichert eine Abhngigkeit dieses Bereichs. Besteht bereits eine Abhngigkeit (gleiche Versionen, gleiche Breiche), so wird die Abhngigkeit nicht erneut
	 * gespeichert.
	 * <p/>
	 * Wurde eine Abhngigkeit gespeichert, die den Typ OPTIONAL {@link ConfigurationAreaDependencyKind} und es soll eine Abhngigkeit REQUIRED gespeichert werden,
	 * so wird die optionale Abhngigkeit ersetzt. Umgekehrt ist dies nicht mglich.
	 * <p/>
	 * <p/>
	 * Ist der Bereich bereits von einem anderen Bereich abhngig (notwendig) und es soll eine Abhngigkeit optional hinzugefgt werden (dessen needed Version >
	 * als die notwendige Abhngigkeit), so wird diese Abhngigkeit automatisch auf "notwendig" gesetzt. Damit wird verhindert, das eine Abhngigkeit, die
	 * notwendig war, durch Versionswechsel wieder auf optional gesetzt werden kann.
	 * <p/>
	 * Ist der Konfiguration die Attributgruppe zum speichern der Datenstze nicht bekannt, so die Abhngigkeiten ignoriert und nichts gemacht.
	 *
	 * @param areaDependencies Abhngigkeiten, die gespeichert werden soll.
	 *
	 * @throws de.bsvrz.dav.daf.main.config.ConfigurationChangeException
	 *          Wird geworfen, wenn der Datensatz, der die Abhngigkeiten enthlt, nicht gespeichert werden kann.
	 */
	synchronized public void addAreaDependency(Collection<ConfigurationAreaDependency> areaDependencies) throws ConfigurationChangeException {

		if(_atguForDependencies != null) {
			/**
			 * Speichert die neuste Abhngigkeit(value) zu einem Bereich (Key). Der neuste Eintrag wird aus den bergebenen Abhngigkeiten rausgesucht.
			 */
			final Map<String, ConfigurationAreaDependency> newestDependence = new HashMap<String, ConfigurationAreaDependency>();

			for(ConfigurationAreaDependency areaDependency : areaDependencies) {
				// Abhngigkeite, die als letztes betrachtet wurde
				final ConfigurationAreaDependency lastFoundDependence = newestDependence.get(areaDependency.getDependantArea());

				if(lastFoundDependence != null) {
					// Enthlt die neue Abhngigkeit neuere Informationen?

					if(hasNewVersionInformation(lastFoundDependence, areaDependency) == true) {
						// Die neue Abhngigkeit enthlt neuere Informationen (hhere Versionsnummern).
						// Also muss die alte Abhngigkeite gelscht werden und durch die neue ersetzt werden. Dabei ist zu beachten, ob die
						// Abhngigkeiten notwendig sind.
						if(lastFoundDependence.getKind() == ConfigurationAreaDependencyKind.REQUIRED) {
							// Die alte Abhngigkeite war notwendig, also sind alle anderen auch notwendig.
							newestDependence.put(areaDependency.getDependantArea(), createRequiredDependence(areaDependency));
						}
						else {
							// Die alte Abhngigkeit war optional, also kann die neue Abhngigkeite die alte einfach ersetzen.
							newestDependence.put(areaDependency.getDependantArea(), areaDependency);
						}
					}
					else if(hasNewStateInformation(lastFoundDependence, areaDependency) == true) {
						// Die Versionen bringen keine neuen Informationen, aber die neue Abhngigkeite war notwendig und die bisher gespeicherte
						// war nur optional. Also muss die bisher gespeicherte Abhngigkeit auf "notwendig" gendert werden
						newestDependence.put(lastFoundDependence.getDependantArea(), createRequiredDependence(lastFoundDependence));
					}
				}
				else {
					// bisher gab es zu dem Bereich noch keine Abhngigkeit
					newestDependence.put(areaDependency.getDependantArea(), areaDependency);
				}
			} // Alle bergebenen Abhngigkeiten nach der Abhngigkeit durchsuchen, die die neuste darstellt

			// An dieser Stelle steht in der Map <code>newestDependence</code> zu jedem abhngigen Bereich die jeweils neuste
			// Abhngigkeite, die bergeben wurde.
			// Nun kann mit der zuletzt gespeicherten Abhngigkeit verglichen werden.

			// Enthlt die Abhngigkeit aus der Map neuere Informationen als die bisher gespeicherte Abhngigkeite, wird
			// die neue Abhngigkeite zustzlich gespeichert.

			final Collection<ConfigurationAreaDependency> possibleNewDependencies = newestDependence.values();

			// Alle bisher gespeicherte Abhngigkeiten werden in _areaDependencyList in der Reihenfolge ihres eintreffens gespeichert.
			// Die Liste muss also vom Ende zum Anfang untersucht werden.

			// Die Ge der Liste muss gespeichert werden, da die neuen Elemente direkt in die Liste eingefgt werden.
			final int oldSize = _areaDependencyList.size() - 1;

			// Diese Liste speichert alle Abhngigkeiten, die an dem bestehenden Datensatz zustzlich gespeichert werden mssen. Alle Versionen und
			// vorallem die Art der Abhngigkeit sind bereits richtig gesetzt.
			final Collection<ConfigurationAreaDependency> saveNewAreaDependencies = new ArrayList<ConfigurationAreaDependency>();

			for(ConfigurationAreaDependency possibleNewDependence : possibleNewDependencies) {
				// Wird bentigt um zu erkennen, dass eine neue Abhngigkeite eingefgt werden muss (bisher gab es zu dem Bereich noch keine Abhngigkeit)
				boolean areaFound = false;
				for(int nr = oldSize; nr >= 0; nr--) {
					// Eine Abhngigkeite, die schon gespeichert wurde
					final ConfigurationAreaDependency savedDependence = _areaDependencyList.get(nr);

					// Wenn die abhngigen Bereiche gleich sind kann verglichen werden.
					if(possibleNewDependence.getDependantArea().equals(savedDependence.getDependantArea())) {
						// Der Bereich stimmt. Da die neusten Abhngigkeiten immer an das Ende der Liste gehngt werden, kann die Suche hier enden.
						areaFound = true;

						// Nun muss geprft werden, ob die neue Abhngigkeit berhaupt neue Informationen beinhaltet.
						if(hasNewVersionInformation(savedDependence, possibleNewDependence) == true) {
							// Wenn die gespeicherte Abhngigkeite notwendig gewesen ist, so muss es auch die neue sein.
							if(savedDependence.getKind() == ConfigurationAreaDependencyKind.REQUIRED) {
								saveNewAreaDependencies.add(createRequiredDependence(possibleNewDependence));
							}
							else {
								// Die Abhngigkeit kann so gespeichert werden wie sie ist.
								saveNewAreaDependencies.add(possibleNewDependence);
							}
						}
						else if(hasNewStateInformation(savedDependence, possibleNewDependence) == true) {
							// Die Versionen unterscheiden sich nicht, aber die alte Abhngigkeite war optional und die neue ist
							// notwendig. Die alte Abhngigkeite wird nicht gendert.
							saveNewAreaDependencies.add(possibleNewDependence);
						}// Im else-Fall muss nichts gemacht werden, die neue Abhngigkeite bringt keine neuen Informationen und muss somit nicht gespeichert werden.

						break;
					}
				}// alle gespeicherten Bereiche betrachten

				if(areaFound == false) {
					// Es gab zu dem Bereich noch keine Abhngigkeit. Dies ist die erste Abhngigkeit zu dem Bereich.
					saveNewAreaDependencies.add(possibleNewDependence);
				}
			}// Prfung, ob die gefundenen Abhngigkeiten berhaupt gespeichert werden mssen.

			// Die gefundenen Abhgigkeiten in die Datenstruktur eintragen und dann an den bestehnden Datensatz anhngen.
			_areaDependencyList.addAll(saveNewAreaDependencies);
			writeDependencyDataSet(saveNewAreaDependencies);
		}// Wird eine alte Konfiguration ohne die ATG benutzt, wird nichts gemacht. Da der Datensatz nicht geschrieben werden kann.
	}

	public boolean dependenciesChecked() {
		if(_atguForDependencies != null) {
			return getConfigurationData(_atguForDependencies) != null;
		}
		else {
			// Dieser Fallunterschied ist eigentlich nicht ntig, da getConfigurationData() null zurck gibt, wenn null bergeben wird.
			// Aber diese Implementierung garantiert ein gleichbleibendes Verhalten.
			return false;
		}
	}

	/**
	 * Schreibt einen Datensatz, der alle Abhngigkeiten dieses Bereich zu anderen Bereichen speichert. Gibt es bereits einen Datensatz, wird dieser um die neuen
	 * Abhngigkeiten erweitert. Die neuen Abhngigkeiten werden ans Ende des Arrays geschrieben.
	 *
	 * @param areaDependencies Neue Abhngigkeiten, die am Ende des Datensatzes (Array) eingefgt werden sollen.
	 *
	 * @throws de.bsvrz.dav.daf.main.config.ConfigurationChangeException
	 *          Der Datensatz darf nicht gendert werden, weil der Konfigurationsverantwortliche der Konfiguration keinen Zugriff auf diesen Bereich hat.
	 */
	synchronized private void writeDependencyDataSet(final Collection<ConfigurationAreaDependency> areaDependencies) throws ConfigurationChangeException {

		if(_atguForDependencies != null) {

			// Dies verhindert, dass der Datensatz erneut geschrieben wird, obwohl es keine neuen Informationen gibt.
			if(areaDependencies.isEmpty() == true) {
				return;
			}

			// Datensatz, der bisher die Abhngigkeiten enthlt
			Data dataSetWithDependencies = getConfigurationData(_atguForDependencies);

			if(dataSetWithDependencies == null) {
				// Der Datensatz wurde bisher noch nicht gesetzt. Die erste Abhngigkeit soll angelegt werden. Damit muss ein Datensatz erzeugt werden.
				final Data newDependancyDataSet = AttributeBaseValueDataFactory.createAdapter(
						_atguForDependencies.getAttributeGroup(), AttributeHelper.getAttributesValues(_atguForDependencies.getAttributeGroup())
				);

				setConfigurationData(_atguForDependencies, newDependancyDataSet);
				// Den Datensatz vom Objekt anfordern
				dataSetWithDependencies = getConfigurationData(_atguForDependencies);
			} // Neuen Datensatz anlegen ende

			// Der bestehende Datensatz muss um die neuen Eintrge erweitert werden.
			final Data.Array dependencyArray = dataSetWithDependencies.getItem("KonfigurationsAbhngigkeiten").asArray();
			final int oldArrayLength = dependencyArray.getLength();
			dependencyArray.setLength(oldArrayLength + areaDependencies.size());

			// Speichert die Stelle, an der der nchste Datensatz eingefgt werden soll.
			int indexCounter = oldArrayLength;

			for(ConfigurationAreaDependency dependence : areaDependencies) {
				// Eine Abhngigkeit, die im Datensatz gespeichert wird
				final Data oneDependencyInDataSet = dependencyArray.getItem(indexCounter);

				oneDependencyInDataSet.getItem("AbhngigkeitEntstandenInVersion").asUnscaledValue().set(dependence.getDependencyOccurredAtVersion());
				oneDependencyInDataSet.getItem("BereichNotwendig").asTextValue().setText(dependence.getDependantArea());
				oneDependencyInDataSet.getItem("VersionNotwendig").asUnscaledValue().set(dependence.getNeededVersion());
				String value = dependence.getKind().getValue();
				value = value.toLowerCase();
				value = value.trim();
				oneDependencyInDataSet.getItem("Kennung").asTextValue().setText(value);

				// Der Wert wurde eingetragen
				indexCounter++;
			}
			setConfigurationData(_atguForDependencies, dataSetWithDependencies);
		} // Wenn eine alte Konfiguration benutzt wird, mache nix
	}

	/**
	 * Diese Methode ldt alle Abhngigkeiten aus dem Datensatz <code>_atguForDependancies</code> in die<code>_areaDependencyList</code>.
	 * <p/>
	 * Der Datensatz speichert die Abhngigkeiten in der Reihenfolge LIFO. Der letzte und damit neuste Eintrag befindet sich am Ende.
	 */
	private void loadDependencyDataSets() {

		if(_atguForDependencies != null) {
			final Data dataWithDependencies = getConfigurationData(_atguForDependencies);
			// Gibt es keinen Datensatz, wird nichts gemacht. Sobald es Abhngigkeiten gibt, wird der Datensatz automatisch angelegt.
			if(dataWithDependencies != null) {
				final Data.Array allDependenciesInDataSets = dataWithDependencies.getItem("KonfigurationsAbhngigkeiten").asArray();

				int index = 0;

				for(int nr = 0; nr < allDependenciesInDataSets.getLength(); nr++) {
					final Data oneDependencyDataSet = allDependenciesInDataSets.getItem(index);
					index++;
					final short dependencyOccuredAt = oneDependencyDataSet.getItem("AbhngigkeitEntstandenInVersion").asUnscaledValue().shortValue();

					final String pidDependencyArea = oneDependencyDataSet.getItem("BereichNotwendig").asTextValue().getValueText();

					final short neededVersion = oneDependencyDataSet.getItem("VersionNotwendig").asUnscaledValue().shortValue();
					final ConfigurationAreaDependencyKind configurationAreaDependencyKind = ConfigurationAreaDependencyKind.getInstance(
							oneDependencyDataSet.getItem(
									"Kennung"
							).asTextValue().getText()
					);

					final ConfigurationAreaDependency areaDependency = new ConfigurationAreaDependency(
							dependencyOccuredAt, neededVersion, pidDependencyArea, configurationAreaDependencyKind
					);
					_areaDependencyList.add(areaDependency);
				}
			}
		} // Wenn eine alte Konfiguration benutzt wird, mache nix
	}

	/**
	 * Prft, ob zwei Abhngigkeiten identisch sind. Welche Werte genau geprft werden, ist von den bergabeparametern abhngig.
	 *
	 * @param first  Erster Bereich
	 * @param second Zweiter Bereich
	 *
	 * @return <code>true</code>, wenn die beiden Bereiche unter Bercksichtigung der Parameter logisch identisch sind.
	 */
	private boolean equalsConfigurationAreaDependancy(final ConfigurationAreaDependency first, ConfigurationAreaDependency second) {
		if(first == second) return true;

		if(first.getDependencyOccurredAtVersion() != second.getDependencyOccurredAtVersion()) {
			return false;
		}

		if(first.getDependantArea().equals(second.getDependantArea())) {
			return false;
		}

		if(first.getNeededVersion() != second.getNeededVersion()) {
			return false;
		}

		if(first.getKind().getCode() != second.getKind().getCode()) {
			return false;
		}

		return true;
	}

	/**
	 * Diese Methode gibt alle eingetragenen Konfigurationsnderungen zurck, die in der Versorgungsdatei eingetragen wurden.
	 *
	 * @return Alle eingetragenen nderungen oder eine leere Collection, falls keine Eintrge gemacht wurden.
	 */
	public Collection<ConfigurationAreaChangeInformation> getChangeLogs() {

		final List<ConfigurationAreaChangeInformation> logs = new ArrayList<ConfigurationAreaChangeInformation>();

		final AttributeGroup atg = _dataModel.getAttributeGroup("atg.konfigurationsnderungen");
		Data data = _dataModel.getConfigurationArea(getPid()).getConfigurationData(atg);
		if(data != null) {
			// Data auslesen und Array erzeugen
			final Data.Array array = data.getArray("Konfigurationsnderung");
			for(int i = 0; i < array.getLength(); i++) {
				Data item = array.getItem(i);
				logs.add(
						new ConfigurationAreaChangeInformation(
								item.getTimeValue("Stand").getMillis(),
								item.getUnscaledValue("Version").intValue(),
								item.getTextValue("Autor").getText(),
								item.getTextValue("Grund").getText(),
								item.getTextValue("Text").getText()
						)
				);
			}
		}

		return logs;
	}

	/**
	 * Erzeugt ein neues dynamisches System-Objekt eines vorgegebenen Typs mit einer angegebenen PID. Optional kann der Name des neuen Objekts vorgegeben werden.
	 * Das neue Objekt wird sofort gltig. Zustzlich knnen alle Datenstze bergeben werden.
	 *
	 * @param type              Typ des neuen Objekts
	 * @param pid               PID des neuen Objekts. Der leere String ("") oder <code>null</code> wird als "keine PID" interpretiert.
	 * @param name              Name des neuen Objekts (kann spter verndert werden)
	 * @param data              Datenstze fr das dynamische Objekt
	 * @param simulationVariant Simulationsvariante des neuen Objekts
	 *
	 * @return Stellvertreterobjekt fr das neu angelegte dynamische Objekt.
	 *
	 * @throws ConfigurationChangeException Wenn das Objekt nicht erzeugt werden konnte.
	 */
	public DynamicObject createDynamicObject(
			DynamicObjectType type, String pid, String name, Collection<DataAndATGUsageInformation> data, short simulationVariant
	) throws ConfigurationChangeException {
		return createDynamicObject(type, pid, name, data, simulationVariant, true);
	}

	/**
	 * Prft, ob ein Name lnger als 255 Zeichen ist.
	 *
	 * @param name Name oder <code>null</code>
	 *
	 * @throws de.bsvrz.dav.daf.main.config.ConfigurationChangeException Der Name ist lnger als 255 Zeichen
	 */
	private void checkNameLength(final String name) throws ConfigurationChangeException {
		if(name != null) {
			if(name.length() <= 255) {
				return;
			}
			else {
				throw new ConfigurationChangeException("Der Name ist lnger als 255 Zeichen " + name + " Lnge " + name.length());
			}
		}
	}

	/**
	 * Prft, ob die Pid lnger als 255 Zeichen ist.
	 *
	 * @param pid Pid oder <code>null</code>
	 *
	 * @throws ConfigurationChangeException Die Pid ist lnger als 255 Zeichen
	 */
	private void checkPidLength(final String pid) throws ConfigurationChangeException {
		if(pid != null) {
			if(pid.length() <= 255) {
				return;
			}
			else {
				throw new ConfigurationChangeException("Die Pid ist lnger als 255 Zeichen " + pid + " Lnge " + pid.length());
			}
		}
	}

	/**
	 * Erzeugt ein neues dynamisches System-Objekt eines vorgegebenen Typs mit einer angegebenen PID. Optional kann der Name des neuen Objekts vorgegeben werden.
	 * Das neue Objekt wird sofort gltig. Zustzlich knnen alle Datenstze bergeben werden.
	 *
	 * @param type              Typ des neuen Objekts
	 * @param pid               PID des neuen Objekts. Der leere String ("") oder <code>null</code> wird als "keine PID" interpretiert.
	 * @param name              Name des neuen Objekts (kann spter verndert werden)
	 * @param data              Datenstze fr das dynamische Objekt
	 * @param simulationVariant Simulationsvariante des neuen Objekts
	 * @param checkDatasets     gibt an, ob geprft werden soll, ob die notwendigen Datenstze vorhanden sind
	 *
	 * @return Stellvertreterobjekt fr das neu angelegte dynamische Objekt.
	 *
	 * @throws ConfigurationChangeException Wenn das Objekt nicht erzeugt werden konnte.
	 */
	private DynamicObject createDynamicObject(
			final DynamicObjectType type,
			String pid,
			String name,
			final Collection<DataAndATGUsageInformation> data,
			final short simulationVariant,
			final boolean checkDatasets
	) throws ConfigurationChangeException {
		// wenn bei der pid oder dem namen "null" bergeben wird, wird der Leerstring ("") verwendet.
		if(pid == null) pid = "";
		if(name == null) name = "";
		if(checkChangePermit()) {

			// Namen und Pids drfen nur eine Lnge von 255 Zeichen besitzen
			checkNameLength(name);
			checkPidLength(pid);

			// gibt es bereits ein aktives Objekt mit der angegebenen Pid, so wird kein neues Objekt angelegt, sondern der Vorgang mit einer
			// ConfigurationChangeException abgebrochen
			SystemObject anotherObject = getDataModel().getObject(pid, simulationVariant);
			if(anotherObject != null && anotherObject.isValid()) {
				throw new ConfigurationChangeException(
						"Es konnte kein neues dynamisches Objekt angelegt werden, da zu der angegebenen Pid '" + pid
						+ "' bereits ein aktives Objekt im Bereich " + getNameOrPidOrId() + " existiert."
				);
			}

			// Alle am Typ definierten Attributgruppen

			final Set<AttributeGroupUsage> requiredAtgUsages = new HashSet<AttributeGroupUsage>();
			if(data != null && !data.isEmpty()) {
				for(DataAndATGUsageInformation dataAndATGUsageInformation : data) {
					final AttributeGroupUsage atgUsage = dataAndATGUsageInformation.getAttributeGroupUsage();
					if(atgUsage.getUsage() == AttributeGroupUsage.Usage.ChangeableRequiredConfigurationData
					   || atgUsage.getUsage() == AttributeGroupUsage.Usage.RequiredConfigurationData) {
						// Erforderluche ATGU merken
						requiredAtgUsages.add(atgUsage);
					}

					// Prfen, ob die Attributgruppe am Typ definiert ist
					if(type instanceof ConfigDynamicObjectType) {
						ConfigDynamicObjectType objectType = (ConfigDynamicObjectType) type;
						objectType.validateAttributeGroup(atgUsage.getAttributeGroup());
					}
				}
			}
			// prfen, ob auch alle notwendigen Datenstze am dynamischen Objekt vorhanden sind
			for(AttributeGroup attributeGroup : type.getAttributeGroups()) {
				for(AttributeGroupUsage atgUsage : attributeGroup.getAttributeGroupUsages()) {
					if(atgUsage.getUsage() == AttributeGroupUsage.Usage.ChangeableRequiredConfigurationData
					   || atgUsage.getUsage() == AttributeGroupUsage.Usage.RequiredConfigurationData) {
						if(!requiredAtgUsages.contains(atgUsage)) {
							final String message = "Es wurden nicht alle notwendigen Datenstze fr das dynamische Objekt " + pid + " angegeben.";
							if(checkDatasets) {
								_debug.error(message);
								throw new ConfigurationChangeException(message);
							}
							else {
								_debug.warning(message);
							}
						}
					}
				}
			}

			// neues Objekt wird angelegt
			final ConfigDataModel configDataModel = getDataModel();
			long nextObjectId = configDataModel.getNextObjectId();
			final ConfigurationAreaFile areaFile = configDataModel.getConfigurationFileManager().getAreaFile(getPid());

			final DynamicObjectInfo objectInfo = areaFile.createDynamicObject(
					nextObjectId, type.getId(), pid, simulationVariant, name, type.getPersistenceMode()
			);
			setTimeOfLastChanges(KindOfLastChange.DynamicObject);


			if(data != null && !data.isEmpty()) {
				// Datenstze am dynamischen Objekt speichern
				for(DataAndATGUsageInformation dataAndATGUsageInformation : data) {
					final AttributeGroupUsage atgUsage = dataAndATGUsageInformation.getAttributeGroupUsage();
					try {
						final Data configurationData = dataAndATGUsageInformation.getData();
						byte[] bytes = new byte[0];
						if(configurationData != null) {
							_dataModel.verifyDataReferences(simulationVariant, configurationData);

							final ByteArrayOutputStream out = new ByteArrayOutputStream();
							Serializer serializer = SerializingFactory.createSerializer(getSerializerVersion(), out);
							serializer.writeData(configurationData);
							bytes = out.toByteArray();
						}
						objectInfo.setConfigurationData(atgUsage.getId(), bytes);
						setTimeOfLastChanges(KindOfLastChange.ConfigurationData);
					}
					catch(Exception ex) {
						final String errorMessage = "Der Datensatz '" + data + "' am Objekt " + getNameOrPidOrId() + " mit der Attributgruppe "
						                            + atgUsage.getAttributeGroup().getNameOrPidOrId() + " und dem Aspekt "
						                            + atgUsage.getAspect().getNameOrPidOrId() + " konnte nicht erstellt werden";
						_debug.error(errorMessage, ex);
						throw new ConfigurationChangeException(errorMessage, ex);
					}
				}
			}

			// beim createSystemObject fr dynamische Objekte wird (im Konstruktor von ConfigDynamicObject) der Typ benachrichtigt,
			// dass ein neues Objekt angelegt wurde
			final DynamicObject dynamicObject = (DynamicObject)configDataModel.createSystemObject(objectInfo);
			_debug.fine("Neues dynamisches Objekt angelegt: " + dynamicObject.getPidOrNameOrId() + " Id: " + dynamicObject.getId());
			((ConfigDynamicObjectType)type).informCreateListener(dynamicObject, simulationVariant);
			return dynamicObject;
		}
		else {
			throw new ConfigurationChangeException(
					"Es konnte kein neues dynamisches Objekt fr diesen Bereich " + getNameOrPidOrId()
					+ " angelegt werden, da die Konfiguration keine Berechtigung hierfr hat."
			);
		}
	}


	/**
	 * Speichert die angegebenen Mengen an einem Konfigurationsobjekt.
	 *
	 * @param objectInfo        wo der konfigurierende Datensatz mit den Mengen abgespeichert werden soll
	 * @param sets              die Mengen
	 * @param serializerVersion die Version des Serialisierers
	 */
	private void createConfigurationDataForSets(ConfigurationObjectInfo objectInfo, Collection<? extends ObjectSet> sets, int serializerVersion) {
		try {
			final ByteArrayOutputStream out = new ByteArrayOutputStream();
			final Serializer serializer = SerializingFactory.createSerializer(serializerVersion, out);
			for(ObjectSet objectSet : sets) {
				serializer.writeLong(objectSet.getId());
			}
			objectInfo.setConfigurationData(-1, out.toByteArray()); // die ATGV "-1" ist fr Mengen fest implementiert
			setTimeOfLastChanges(KindOfLastChange.ConfigurationData);
			out.close();
		}
		catch(Exception ex) {
			final String errorMessage = "Konfigurierender Datensatz konnte nicht geschrieben werden, da es Probleme mit dem Serializer gibt";
			_debug.error(errorMessage, ex);
			throw new RuntimeException(errorMessage, ex);
		}
	}

	/**
	 * Gibt die Versionsnummer dieses Konfigurationsbereichs zurck, die zur angegebenen Zeit aktiv war.
	 *
	 * @param time die Zeit, zu der die Versionsnummer des Konfigurationsbereichs gesucht wird
	 *
	 * @return die Versionsnummer dieses Konfigurationsbereichs zur angegebenen Zeit oder <code>0</code>, falls der Konfigurationsbereich noch nie aktiviert wurde
	 */
	short getVersionAtAssignedTime(long time) {
		final List<VersionInfo> versionInfos = getDataModel().getVersionInfoOfConfigurationArea(getConfigurationArea());
		for(int i = versionInfos.size() - 1; i >= 0; i--) {	// Schleife durchluft die Liste rckwrts
			VersionInfo versionInfo = versionInfos.get(i);
			if(versionInfo.getActivationTime() <= time) return versionInfo.getVersion();
		}
		return 0;
	}

	/**
	 * Gibt die Aktivierungszeit dieses Konfigurationsbereichs zur angegebenen Versionsnummer zurck.
	 *
	 * @param version die Versionsnummer, dessen Aktivierungszeit gewnscht ist
	 *
	 * @return die Aktivierungszeit der angegebenen Version dieses Konfigurationsbereichs oder <code>0</code>, falls zur Version kein Eintrag vorliegt
	 */
	long getTimeAtAssignedVersion(short version) {
		final List<VersionInfo> versionInfos = getDataModel().getVersionInfoOfConfigurationArea(getConfigurationArea());
		for(VersionInfo versionInfo : versionInfos) {
			if(versionInfo.getVersion() == version) return versionInfo.getActivationTime();
		}
		return 0;
	}

	/**
	 * Wurde ein dynamisches Objekt, ein Konfigurationsobjekt (welches noch nicht aktiviert ist) oder ein nderbarer konfigurierender Datensatz gendert, so wird
	 * die nderungszeit in dem Attribut des Datensatzes eingetragen und abgespeichert.
	 *
	 * @param kind gibt an, welche der drei Zeiten angepasst werden muss
	 */
	void setTimeOfLastChanges(KindOfLastChange kind) {
		try {
			long time = System.currentTimeMillis();
			// Datensatz holen und wegschreiben
			long timeOfDynamicObject = 0;
			long timeOfConfigurationObject = 0;
			long timeOfConfigurationData = 0;

			// bisherige Daten werden aus dem Byte-Array eingelesen, da ich kein Data habe, wenn noch kein Datensatz existiert.
			final AttributeGroup atg = getDataModel().getAttributeGroup("atg.konfigurationsBereichnderungsZeiten");
			final Aspect asp = getDataModel().getAspect("asp.eigenschaften");
			final AttributeGroupUsage atgUsage = atg.getAttributeGroupUsage(asp);
			byte[] bytes = getConfigurationDataBytes(atgUsage);
			if(bytes != null) {
				final ByteArrayInputStream in = new ByteArrayInputStream(bytes);	// falls Daten vorhanden sind werden sie eingelesen
				final Deserializer deserializer = SerializingFactory.createDeserializer(getSerializerVersion(), in);
				timeOfDynamicObject = deserializer.readLong();
				timeOfConfigurationObject = deserializer.readLong();
				timeOfConfigurationData = deserializer.readLong();
				in.close();
			}

			// welche Zeit hat sich gendert?
			if(kind == KindOfLastChange.DynamicObject) {
				timeOfDynamicObject = time;
			}
			else if(kind == KindOfLastChange.ConfigurationObject) {
				timeOfConfigurationObject = time;
			}
			else if(kind == KindOfLastChange.ConfigurationData) {
				timeOfConfigurationData = time;
			}
			else {
				throw new RuntimeException(
						"Die Zeiten der letzten nderung eines Bereichs knnen nicht verndert werden, da die angegebene Art '" + kind + "'unbekannt ist."
				);
			}

			// neue Daten werden ins Byte-Array geschrieben
			final ByteArrayOutputStream out = new ByteArrayOutputStream();
			final Serializer serializer = SerializingFactory.createSerializer(getSerializerVersion(), out);
			serializer.writeLong(timeOfDynamicObject);
			serializer.writeLong(timeOfConfigurationObject);
			serializer.writeLong(timeOfConfigurationData);
			_systemObjectInfo.setConfigurationData(atgUsage.getId(), out.toByteArray());
			out.close();
		}
		catch(Exception ex) {
			final String errorMessage = "Der Datensatz mit der letzten nderung eines Objekts oder Datensatzes konnte am Bereich " + getNameOrPidOrId()
			                            + " nicht geschrieben werden.";
			_debug.error(errorMessage, ex);
			throw new RuntimeException(errorMessage, ex);
		}
	}

	/**
	 * Prft, ob es einen Wert im Datensatz <code>atg.konfigurationsBereichnderungsZeiten</code> gibt, der <code>Undefiniert (0)</code> ist, und ersetzt diesen
	 * mit der aktuellen Zeit.
	 */
	public void initialiseTimeOfLastChanges() {
		if(getTimeOfLastDynamicChange() == 0) setTimeOfLastChanges(KindOfLastChange.DynamicObject);
		if(getTimeOfLastActiveConfigurationChange() == 0) setTimeOfLastChanges(KindOfLastChange.ConfigurationData);
		if(getTimeOfLastNonActiveConfigurationChange() == 0) setTimeOfLastChanges(KindOfLastChange.ConfigurationObject);
	}
}
