/*
 * Copyright 2013 by Kappich Systemberatung Aachen
 * 
 * This file is part of de.bsvrz.puk.config.
 * 
 * de.bsvrz.puk.config is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.puk.config is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.puk.config; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package de.bsvrz.puk.config.main.consistencycheck;

import de.bsvrz.dav.daf.communication.dataRepresentation.AttributeBaseValueDataFactory;
import de.bsvrz.dav.daf.communication.dataRepresentation.AttributeHelper;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.puk.config.configFile.datamodel.ConfigNonMutableSet;
import de.bsvrz.puk.config.configFile.datamodel.ConfigSystemObject;
import de.bsvrz.puk.config.xmlFile.properties.ConfigurationValueRange;
import de.bsvrz.sys.funclib.debug.Debug;

import java.util.Arrays;
import java.util.Set;
import java.util.TreeSet;

/**
 * Hilfsklasse fr die Spezialbehandlungen bei unversionierten Datenmodellnderungen.
 *
 * Der Import luft in zwei Schritten ab. Zurerst wird mit den Processable-Methoden geprft, ob unversionierte nderungen
 * mglich sind (true) oder nicht (false). Wenn unversionierte nderungen mglich sind und auch keine anderen nderungen dagegen sprechen wird
 * spter dann bei den entsprechenden allow...-Methoden nocheinmal geprft, ob die nderugnen durchgefhrt werden drfen (die Logik muss
 * natrlich identisch sein). Da die allow...-Methoden direkt vor dem ndern der Konfigurationsdateien ausgefhrt werden, wird in den allow...-
 * Methoden auch der entsprechende Datensatz am Konfigurationsbereich ergnzt, der Informationen ber unversionierte nderungen enthlt.
 *
 * Dieser Datensatz darf noch nicht bei den Processable-Methoden ergnzt werden, da dort noch niocht feststeht, dass nicht auch andere
 * Aktionen durchgefhrt werden, die trotzdem eine neue Objektversion bewirken wrden (z.B. hinzufgen vo nStatuswerten (wrde erlaubt werden)
 * mit gleichzeitigem Lschen von Statuswerten.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 11583 $
 */
public class RelaxedModelChanges {

	private static final Debug _debug = Debug.getLogger();

	private static boolean _isEnabled;
	private final AttributeGroup _attributeGroup;

	public RelaxedModelChanges(final DataModel dataModel) {
		_attributeGroup = dataModel.getAttributeGroup("atg.konfigurationsBereichUnversioniertenderungen");
		if(_attributeGroup == null){
			_debug.fine(
					"Unversionierte nderungen nicht mglich, atg.konfigurationsBereichUnversioniertenderungen nicht gefunden. Bitte kb.metaModellGlobal in Mindestversion 16 installieren."
			);
			_isEnabled = false;
		}
		else {
			_debug.fine(
					"Unversionierte nderungen aktiviert."
			);
			_isEnabled = true;
		}
	}

	/**
	 * Sonderbehandlung fr Wertebereiche. Das ndern, wird erlaubt, wenn der Bereich nicht verkleinert wird und Einheit und Skalierung beibehalten werden.
	 * @param importedValueRange Eingelesene Bereichsdefinition
	 * @param existingValueRange Existierende Bereichsdefinition
	 * @return true: die nderung wird erlaubt, false: keine Sonderbehandlung, hier nicht erlaubt
	 */
	public boolean isValueRangeChangeProcessable(final ConfigurationValueRange importedValueRange, final IntegerValueRange existingValueRange) {

		if(!_isEnabled) return false;

		// Skalierung prfen
		if(importedValueRange.getScale() != existingValueRange.getConversionFactor()) return false;

		// Einheit prfen
		if(!importedValueRange.getUnit().equals(existingValueRange.getUnit())) return false;

		// Bereich prfen
		if(importedValueRange.getMinimum() > existingValueRange.getMinimum()) return false;
		if(importedValueRange.getMaximum() < existingValueRange.getMaximum()) return false;

		return true;
	}
	/**
	 * Gibt zurck, ob zu dem angegebenen IntegerAttributeType neue Statuswerte hinzugefgt werden drfen
	 * @param integerAttributeType IntegerAttributeType
	 * @return true falls neue Statuswerte hinzugefgt werden drfen, sonst false
	 */
	public boolean isAddStatesProcessable(final IntegerAttributeType integerAttributeType) {
		if(!_isEnabled) return false;
		return true;
	}

	/**
	 * Gibt zurck, ob die maximale Lnge eines Feldes gendert werden darf. Vorraussetzung: oldMaxCount != newMaxCount und das
	 * Feld hat eine variable Lnge.
	 * @param oldMaxCount Alte Lnge
	 * @param newMaxCount Neue Lnge
	 * @return Darf die Feldlnge ohne Versionierung gendert werden?
	 */
	public boolean isChangeArrayMaxCountProcessable(final int oldMaxCount, final int newMaxCount) {
		if(!_isEnabled) return false;
		return newMaxCount >= oldMaxCount && getDataType(oldMaxCount) == getDataType(newMaxCount);
	}

	/**
	 * Gibt zurck, ob das ndern eines Wert-Namens erlaubt sein soll, ohne eine neue Version zu erstellen.
	 * @param integerValueState Statuswert
	 * @return Darf der Name gednert werden?
	 */
	public boolean isChangeValueNameProcessable(final IntegerValueState integerValueState) {
		if(!_isEnabled) return false;

		// Zugehrigen Attributtyp suchen
		IntegerAttributeType integerAttributeType = null;
		for(SystemObject systemObject : integerValueState.getConfigurationArea().getCurrentObjects()) {
			if(systemObject instanceof IntegerAttributeType) {
				IntegerAttributeType tmp = (IntegerAttributeType) systemObject;
				if (tmp.getStates().contains(integerValueState)){
					integerAttributeType = tmp;
				}
			}
		}

		if(integerAttributeType == null) return false;

		return true;
	}

	/**
	 * Gibt true zurck sofern der Wertebereich bearbeitet werden darf und markiert den zugehrigen Attributtyp als durch unversionierte nderungen bearbeitet.
	 *
	 * @param configSystemObject
	 * @param oldData Existierende Bereichsdefinition Alter Datensatz
	 * @param newData Eingelesene Bereichsdefinition Neuer Datensatz
	 * @return true: die nderung wird erlaubt, false: keine Sonderbehandlung, hier nicht erlaubt
	 */
	public boolean allowChangeValueRange(final ConfigSystemObject configSystemObject, final Data oldData, final Data newData) {

		if(!_isEnabled) return false;

		if(oldData == null || newData == null) return false;

		// Zugehrigen Attributtyp suchen
		IntegerAttributeType integerAttributeType = null;
		for(SystemObject systemObject : configSystemObject.getConfigurationArea().getCurrentObjects()) {
			if(systemObject instanceof IntegerAttributeType) {
				IntegerAttributeType tmp = (IntegerAttributeType) systemObject;
				if (tmp.getRange().equals(configSystemObject)){
					integerAttributeType = tmp;
				}
			}
		}

		if(integerAttributeType == null) return false;

		// Skalierung prfen
		if(newData.getScaledValue("skalierung").doubleValue() != oldData.getScaledValue("skalierung").doubleValue()) return false;

		// Einheit prfen
		if(!newData.getTextValue("einheit").getValueText().equals(oldData.getTextValue("einheit").getValueText())) return false;

		// Bereich prfen
		if(newData.getUnscaledValue("minimum").longValue() > oldData.getUnscaledValue("minimum").longValue()) return false;
		if(newData.getUnscaledValue("maximum").longValue() < oldData.getUnscaledValue("maximum").longValue()) return false;

		if(!markAsUnversionedChanged(integerAttributeType)) return false;
		return true;
	}
	/**
	* Gibt zurck, ob das ndern eines Wert-Namens erlaubt sein soll und markiert den zugehrigen Attributtyp als durch unversionierte nderungen bearbeitet.
	* @param integerValueState Statuswert
	* @return Darf der Name gendert werden?
	*/
	public boolean allowChangeValueName(final IntegerValueState integerValueState) {
		if(!_isEnabled) return false;

		// Zugehrigen Attributtyp suchen
		IntegerAttributeType integerAttributeType = null;
		for(SystemObject systemObject : integerValueState.getConfigurationArea().getCurrentObjects()) {
			if(systemObject instanceof IntegerAttributeType) {
				IntegerAttributeType tmp = (IntegerAttributeType) systemObject;
				if (tmp.getStates().contains(integerValueState)){
					integerAttributeType = tmp;
				}
			}
		}

		if(integerAttributeType == null) return false;
		if(!markAsUnversionedChanged(integerAttributeType)) return false;

		return true;
	}


	/**
	 * Gibt true zurck, sofern der Wertezustand hinzugefgt werden darf und markiert den zugehrigen Attributtyp als durch unversionierte nderungen bearbeitet.
	 * @param objectSet Objekt-Menge
	 * @return
	 */
	public boolean allowObjectSetAdd(final ConfigNonMutableSet objectSet) {

		if(!_isEnabled) return false;

		if(!objectSet.getObjectSetType().getPid().equals("menge.werteZustaende")) {
			return false;
		}
		for(SystemObject systemObject : objectSet.getConfigurationArea().getCurrentObjects()) {
			if(systemObject instanceof IntegerAttributeType) {
				IntegerAttributeType integerAttributeType = (IntegerAttributeType) systemObject;
				if (objectSet.equals(integerAttributeType.getNonMutableSet(objectSet.getName()))){
					return markAsUnversionedChanged(integerAttributeType);
				}
			}
		}
		return false;
	}

	/**
	 * Gibt true zurck, falls die Lnge des Daten-Arrays gendert werden darf und markiert den zugehrigen Attributtyp als durch unversionierte nderungen bearbeitet.
	 *
	 * @param configSystemObject SystemObjekt des Attributes
	 * @param oldData Existierende Attributtypeigenschaften Alter Datensatz
	 * @param newData Eingelesene Attributtypeigenschaften Neuer Datensatz
	 * @return true: die nderung wird erlaubt, false: keine Sonderbehandlung, hier nicht erlaubt
	 */
	public boolean allowChangeArrayMaxCount(final ConfigSystemObject configSystemObject, final Data oldData, final Data newData) {
		if(!_isEnabled) return false;

		if(oldData == null || newData == null) return false;

		if(!(configSystemObject instanceof Attribute)) {
			return false;
		}

		Attribute attribute = (Attribute) configSystemObject;
		AttributeType attributeType = attribute.getAttributeType();


		// Position muss bereinstimmen
		if(newData.getScaledValue("position").doubleValue() != oldData.getScaledValue("position").doubleValue()) return false;

		// Anzahl-Variabel muss bereinstimmen...
		if(!newData.getTextValue("anzahlVariabel").getValueText().equals(oldData.getTextValue("anzahlVariabel").getValueText())) return false;
		// ... und auf "Ja" stehen.
		if(!newData.getTextValue("anzahlVariabel").getValueText().equals("Ja")) return false;

		// Attributtyp muss bereinstimmen
		if(!(newData.getReferenceValue("attributTyp").getId() == oldData.getReferenceValue("attributTyp").getId())) return false;

		// Lnge darf nur grer werden, muss aber im selben Wertebereich bleiben
		if(!isChangeArrayMaxCountProcessable(oldData.getUnscaledValue("anzahl").intValue(), newData.getUnscaledValue("anzahl").intValue())) {
			return false;
		}

		if(!markAsUnversionedChanged(attributeType)) return false;

		return true;
	}

	private boolean markAsUnversionedChanged(final AttributeType attributeType) {

		if(!_isEnabled) return false;

		ConfigurationArea configurationArea = attributeType.getConfigurationArea();
		Data configurationData = configurationArea.getConfigurationData(_attributeGroup);
		if(configurationData == null){
			_debug.warning(
					"Fhre unversionierte nderung am Attributtyp " + attributeType + " durch. Der Konfigurationsbereich wird kb.metaModellGlobal in Mindestversion 16 voraussetzen."
			);
			configurationData = AttributeBaseValueDataFactory.createAdapter(_attributeGroup, AttributeHelper.getAttributesValues(_attributeGroup));
		}
		else {
			_debug.fine(
					"Fhre unversionierte nderung am Attributtyp " + attributeType + " durch."
			);
		}
		boolean added = false;
		Data versions = configurationData.getItem("versionen");
		for(Data data : versions) {
			if(data.getUnscaledValue("Version").shortValue() == configurationArea.getModifiableVersion()){
				Data.TextArray array = data.getTextArray("AttributTypen");
				Set<String> systemObjects = new TreeSet<String>(Arrays.asList(array.getTextArray()));
				systemObjects.add(attributeType.getPid());
				array.set(systemObjects.toArray(new String[systemObjects.size()]));
				added = true;
			}
		}
		if(!added){
			Data.Array array = versions.asArray();
			array.setLength(array.getLength() + 1);
			Data last = array.getItem(array.getLength() - 1);
			last.getUnscaledValue("Version").set(configurationArea.getModifiableVersion());
			last.getTextArray("AttributTypen").set(attributeType.getPid());
		}
		try {
			configurationArea.setConfigurationData(_attributeGroup, configurationData);
		}
		catch(ConfigurationChangeException e) {
			_debug.warning("Konnte unversionierte nderung nicht durchfhren. Der Datensatz am Konfigurationsbereich konnte nicht geschrieben werden", e);
			return false;
		}
		return true;
	}

	/**
	 * Gibt fr jeden Wertebereich (Byte, Short, Integer) eine eindeutige Nummer (0,1,2) zurck.
	 * @param maxCount
	 */
	private static int getDataType(final int maxCount) {
		if(maxCount < 0) throw new IllegalArgumentException();
		if(maxCount < 256) return 0;
		if(maxCount < 65536) return 1;
		return 2;
	}

	public static RelaxedModelChanges getInstance(final DataModel dataModel) {
		// Objekt sehr klein, halten der Objekte in einer HashMap o.. vorerst unntig.
		return new RelaxedModelChanges(dataModel);
	}
}
